# NYC Taxi fare prediction sample dataset.
# Ref: https://www1.nyc.gov/site/tlc/about/tlc-trip-record-data.page
NYC_TAXI_SMALL_URL = (
    "https://azurefeathrstorage.blob.core.windows.net/public/sample_data/green_tripdata_2020-04_with_index.csv"
)

# Fraud detection sample datasets.
# Ref: https://github.com/microsoft/r-server-fraud-detection
FRAUD_DETECTION_ACCOUNT_INFO_URL = (
    "https://raw.github.com/microsoft/r-server-fraud-detection/master/Data/Account_Info.csv"
)

FRAUD_DETECTION_FRAUD_TRANSACTIONS_URL = (
    "https://raw.github.com/microsoft/r-server-fraud-detection/master/Data/Fraud_Transactions.csv"
)

FRAUD_DETECTION_UNTAGGED_TRANSACTIONS_URL = (
    "https://raw.github.com/microsoft/r-server-fraud-detection/master/Data/Untagged_Transactions.csv"
)

# Product recommendation sample datasets.
# Ref:
PRODUCT_RECOMMENDATION_USER_OBSERVATION_URL = (
    "https://azurefeathrstorage.blob.core.windows.net/public/sample_data/product_recommendation_sample/user_observation_mock_data.csv"
)

PRODUCT_RECOMMENDATION_USER_PROFILE_URL = (
    "https://azurefeathrstorage.blob.core.windows.net/public/sample_data/product_recommendation_sample/user_profile_mock_data.csv"
)

PRODUCT_RECOMMENDATION_USER_PURCHASE_HISTORY_URL = (
    "https://azurefeathrstorage.blob.core.windows.net/public/sample_data/product_recommendation_sample/user_purchase_history_mock_data.csv"
)

PRODUCT_RECOMMENDATION_PRODUCT_DETAIL_URL = (
    "https://azurefeathrstorage.blob.core.windows.net/public/sample_data/product_recommendation_sample/product_detail_mock_data.csv"
)

# Hotel review sample datasets.
# Ref: https://www.kaggle.com/datasets/datafiniti/hotel-reviews
HOTEL_REVIEWS_URL = (
    "https://raw.githubusercontent.com/Azure-Samples/azure-search-sample-data/main/hotelreviews/HotelReviews_data.csv"
)
