"""Utilities for downloading sample datasets"""

from feathr.datasets.constants import (
    NYC_TAXI_SMALL_URL,
    FRAUD_DETECTION_ACCOUNT_INFO_URL,
    FRAUD_DETECTION_FRAUD_TRANSACTIONS_URL,
    FRAUD_DETECTION_UNTAGGED_TRANSACTIONS_URL,
    PRODUCT_RECOMMENDATION_USER_OBSERVATION_URL,
    PRODUCT_RECOMMENDATION_USER_PROFILE_URL,
    PRODUCT_RECOMMENDATION_USER_PURCHASE_HISTORY_URL,
    PRODUCT_RECOMMENDATION_PRODUCT_DETAIL_URL,
    HOTEL_REVIEWS_URL,
)

__all__ = [
    "NYC_TAXI_SMALL_URL",
    "FRAUD_DETECTION_ACCOUNT_INFO_URL",
    "FRAUD_DETECTION_FRAUD_TRANSACTIONS_URL",
    "FRAUD_DETECTION_UNTAGGED_TRANSACTIONS_URL",
    "PRODUCT_RECOMMENDATION_USER_OBSERVATION_URL",
    "PRODUCT_RECOMMENDATION_USER_PROFILE_URL",
    "PRODUCT_RECOMMENDATION_USER_PURCHASE_HISTORY_URL",
    "PRODUCT_RECOMMENDATION_PRODUCT_DETAIL_URL",
    "HOTEL_REVIEWS_URL",
]
